--************************************************************************
--  serial.adb
--
--  Written by:  C1C Hugh O'Donnell 
--               C1C Robin Stephens
--               Lt. Col. Joiner
--
--  Description:   This program is an example of serial port communication 
--                 via Win32 binding.  Must
--                 be run on a computer running Windoze 95 or Windoze NT.
--
--  Modified by: C1C Andrew Lundquist 
--  Modified on: 7 July 98
--
--  Modifications: 
--
--                 Supports COM1 - COM4 (if your machine has them).
--
--                 The get procedure no longer waits until something is
--                 received - it checks and then returns control to the 
--                 system if nothing is there.               
-- 
--************************************************************************
with System; use System;

with Ada.Unchecked_Conversion;
with Win32;
with Interfaces.C.Strings;
with Win32.Winbase;
with Win32.Winnt;
with Ada.Exceptions;

package body serial is

--************************************************************************
--  Package types:  types that we used to make coding easier
--************************************************************************
   use type Win32.Winnt.Handle;

   use type Win32.Dword;

   use type Win32.Bool;

   type Char_Ptr is access all Interfaces.C.Char;

--************************************************************************
--  Package Variable Port_Handle : This variable is used to hold the
--			necessary Win32 information to reference the port throughout
--			the entire package      
--************************************************************************
   Port_Handle : Win32.Winnt.Handle;


--************************************************************************
--  Package Variable Port_Open : This variable indicates whether the port 
--  		is open (true) or closed (false)
--************************************************************************
   Port_Open : boolean;


--************************************************************************
--  Private Functions: Most of these are unchecked-conversion functions 
--  that allow us to go between types (Thanks go to Lt. Col. Joiner) and
--  the necessary procedures to maipulate a communicaitons port.  
--************************************************************************
   function char_to_byte is new
      Ada.Unchecked_Conversion
      (Source => character,
      Target => byte);

   function byte_to_char is new
      Ada.Unchecked_Conversion
      (Source => byte,
      Target => character);

   function Convert_Char_Ptr_To_Lpvoid is new
      Ada.Unchecked_Conversion
      (Source => Char_Ptr,
      Target => Win32.Lpvoid);

   function Convert_Chars_Ptr_To_Lpcstr is new
      Ada.Unchecked_Conversion
      (Source => Interfaces.C.Strings.Chars_Ptr,
      Target => Win32.Lpcstr);

   function Convert_Chars_Ptr_To_Lpcvoid is new
      Ada.Unchecked_Conversion
      (Source => Interfaces.C.Strings.Chars_Ptr,
      Target => Win32.Lpcvoid);

   function Convert_Pvoid_To_Handle is new
      Ada.Unchecked_Conversion
      (Source => Win32.Pvoid,
      Target => Win32.Winnt.Handle);

   function Convert_Pvoid_To_Lpsecurity_Attributes is new
      Ada.Unchecked_Conversion
      (Source => Win32.Pvoid,
      Target => Win32.Winbase.Lpsecurity_Attributes);

   function Convert_Pvoid_To_Lpoverlapped is new
      Ada.Unchecked_Conversion
      (Source => Win32.Pvoid,
      Target => Win32.Winbase.Lpoverlapped);


--************************************************************************
--  Function:     open
--
--  Parameters:   port : a four letter string that needs to be either 
--                       COM1 or COM2
--
--  Description:  Establishes a connection to the serial port.
--************************************************************************
procedure open (port : in string) is

   nulla       : System.Address := Null_Address;

begin
   port_handle :=
      Win32.Winbase.CreateFile
      (lpFileName => Convert_Chars_Ptr_To_Lpcstr
                     (Interfaces.C.Strings.New_String
                     (Str => port)), -- Win32.Lpcstr
       dwDesiredAccess => Win32.Dword(Win32.Winnt.Generic_Read) or
                          Win32.Dword(Win32.Winnt.Generic_Write), -- Win32.Dword
       dwShareMode => Win32.Dword(0), -- Win32.Dword
       lpSecurityAttributes => Convert_Pvoid_To_Lpsecurity_Attributes
                               (nulla), -- Win32.Winbase.Lpsecurity_Attributes
       dwCreationDisposition => Win32.Dword(Win32.Winbase.Open_Existing), -- Win32.Dword
       dwFlagsAndAttributes => Win32.Dword(0), -- Win32.Dword
       hTemplateFile => Convert_Pvoid_To_Handle(nulla)); -- Win32.Winnt.Handle

--This sets package variable port_open to true indicating the port
--port was successfully opened, or raises an exception
   if Port_Handle /= Win32.Winbase.Invalid_Handle_Value then
      Port_Open := true;
   else
      raise Operation_Failed;
   end if;
end open;


--************************************************************************
--  Function:     close
--
--  Parameters:   None
--
--  Description:  Terminates the connection to the serial port.
--************************************************************************
procedure close is

   Bool_Result : Win32.Bool;

begin
   Bool_Result :=
      Win32.Winbase.CloseHandle
      (hObject => Port_Handle); -- Win32.Winnt.Handle

--This sets the package variable Port_Open to false indicating the port
--was successfully closed, or else raise exception
   if Bool_Result /= 0 then
      Port_Open := false;
   else
      raise Operation_Failed;
   end if;

end close;


--************************************************************************
--  Function:     set_up_port
--
--  Parameters:   Baud : an integer that sets the baud rate (needs to be
--                       one of these numbers:  1200, 2400, 4800, 9600)
--
--  Description:  Sets the port up with predetermined settings.
--
--                Also, set the timeouts for the port.
--************************************************************************
procedure set_up_port(baud : in integer; done : out boolean) is

   Bool_Result : Win32.Bool;
   Dcb : Win32.Winbase.Lpdcb := new Win32.Winbase.Dcb;

   CommTimeouts : Win32.Winbase.Lpcommtimeouts := new Win32.Winbase.Commtimeouts;

begin
   Done := True;
   Bool_Result :=
      Win32.Winbase.GetCommState
      (hFile => Port_Handle,                       --Win32.Winnt.Handle
       lpDCB => Dcb);                              --Win32.Winbase.Lpdcb

   if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "GetCommState returned 0");
   end if;

--Attempting to set the baud rate
   Dcb.all.BaudRate := win32.dword(baud);         --Add baud to DCB
   Bool_Result :=
      Win32.Winbase.SetCommState
      (hFile => Port_Handle,                      --Win32.Winnt.Handle
       lpDCB => Dcb);                             --Win32.Winbase.Lpdcb

   if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "SetCommState (baud rate) returned 0");
      Done := False;
   end if;

   Bool_Result :=
    Win32.Winbase.GetCommTimeouts
     (hFile =>
        Port_Handle, -- Win32.Winnt.Handle
      lpCommTimeouts =>
        CommTimeouts); -- Lpcommtimeouts

  if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "GetCommTimeouts returned 0");
  end if;

  CommTimeouts.all.ReadIntervalTimeout        := 250;
  CommTimeouts.all.ReadTotalTimeoutMultiplier := 250;
  CommTimeouts.all.ReadTotalTimeoutConstant   := 250;

  Bool_Result :=
    Win32.Winbase.SetCommTimeouts
     (hFile =>
        Port_Handle, -- Win32.Winnt.Handle
      lpCommTimeouts =>
        CommTimeouts); -- Lpcommtimeouts

  if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "SetCommTimeouts returned 0");
  end if;

end set_up_port;

--************************************************************************
--  Function:     get
--
--  Parameters:   Item : A byte of data that is received at the serial port
--
--  Description:  Reads an item from the serial port.
--************************************************************************
procedure get(item : out byte) is

   char_item : character;
   Bool_Result : Win32.Bool;
   NumberOfBytesRead : Win32.Lpdword := new Win32.Dword;
   Char_Buffer : Char_Ptr := new Interfaces.C.Char;

   nulla       : System.Address := Null_Address;

begin
--Read in a byte of information
   Bool_Result :=
      Win32.Winbase.ReadFile
      (hFile => Port_Handle,                                -- Win32.Winnt.Handle
       lpBuffer => Convert_Char_Ptr_To_Lpvoid(Char_Buffer), -- Win32.Lpvoid
       nNumberOfBytesToRead => Win32.Dword(1),              -- Win32.Dword
       lpNumberOfBytesRead => NumberOfBytesRead,            -- Win32.Lpdword
       lpOverlapped => Convert_Pvoid_To_Lpoverlapped(nulla));-- Lpoverlapped

   if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "ReadFile returned 0");
   end if;

--This sets the exportable character type from the character buffer
--*****NOTE: Need to change this to a 'byte' type*****
   char_item := Interfaces.C.To_Ada(Char_Buffer.all);
   item := char_to_byte(char_item);

end get;


--************************************************************************
--  Function:     put
--
--  Parameters:   Item : A byte of data to put to the serial port
--
--  Description:  Writes an item to the serial port.
--************************************************************************
procedure put(item : in byte) is

   char_item : character;
   Bool_Result : Win32.Bool;
   NumberOfBytesWritten : Win32.Lpdword := new Win32.Dword;

   nulla       : System.Address := Null_Address;

begin
   char_item := byte_to_char(item);

--Put a byte to the port
   Bool_Result :=
      Win32.Winbase.WriteFile
      (hFile => Port_Handle,                                --Win32.Winnt.Handle
       lpBuffer => Convert_Chars_Ptr_To_Lpcvoid
                   (Interfaces.C.Strings.New_String
                   (Str => char_item & Ascii.Cr)),          --Win32.Lpcvoid
       nNumberOfBytesToWrite => Win32.Dword(1),             -- Win32.Dword  => 1 bytes
       lpNumberOfBytesWritten => NumberOfBytesWritten,      -- Win32.Lpdword
       lpOverlapped => Convert_Pvoid_To_Lpoverlapped(nulla));-- Lpoverlapped

   if Bool_Result = 0 then
      Ada.Exceptions.Raise_Exception(Operation_Failed'Identity,
         "WriteFile returned 0");
   end if;
end put;


--************************************************************************
--  Function:     port_status
--
--  Parameters:   None
--
--  Description:  Returns whether the port is ready to be written/read.
--************************************************************************
function  port_status return boolean is

begin
   return Port_Open;
end;

end serial;
